import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
import pytest
from sqlalchemy_query_helper.query_generator import generate_query
from tests.model import Base, User

engine = create_engine("sqlite:///:memory:", echo=True)
Session = sessionmaker(bind=engine)


@pytest.fixture(scope="function", name="session_fixture")
def _session_fixture():
    Base.metadata.create_all(engine)
    session = Session()
    yield session
    session.close()
    Base.metadata.drop_all(bind=engine)


def test_eq_operator(session_fixture):
    ed_user1 = User(name="ed", fullname="Ed Jones", nickname="edsnickname")
    ed_user2 = User(name="eds", fullname="Ed Jones", nickname="edsnickname")
    session_fixture.add(ed_user1)
    session_fixture.add(ed_user2)
    query = generate_query(
        session_fixture, User, {"name": {"op": "eq", "value": ed_user1.name}}
    )
    result = list(query)
    assert len(result) == 1
    assert result[0].name == ed_user1.name


def test_multiple_operator(session_fixture):
    ed_user1 = User(name="ed", fullname="Ed Jones", nickname="edsnickname")
    ed_user2 = User(name="ed", fullname="Ed Joness", nickname="edsnickname")
    session_fixture.add(ed_user1)
    session_fixture.add(ed_user2)
    query = generate_query(
        session_fixture,
        User,
        {
            "name": {"op": "eq", "value": ed_user2.name},
            "fullname": {"op": "neq", "value": ed_user1.fullname},
        },
    )
    result = list(query)
    assert len(result) == 1
    assert result[0].name == ed_user2.name
    assert result[0].fullname == ed_user2.fullname


def test_gte_with_datetime(session_fixture):
    ed_user1 = User(
        name="ed",
        fullname="Ed Jones",
        nickname="edsnickname",
        timestamp=datetime.datetime.utcnow() + datetime.timedelta(days=1),
    )
    ed_user2 = User(
        name="ed",
        fullname="Ed Jones",
        nickname="edsnickname",
        timestamp=datetime.datetime.utcnow() - datetime.timedelta(days=1),
    )
    session_fixture.add(ed_user1)
    session_fixture.add(ed_user2)
    query = generate_query(
        session_fixture,
        User,
        {"timestamp": {"op": "gte", "value": ed_user1.timestamp}},
    )
    result = list(query)
    assert len(result) == 1
    assert result[0].timestamp >= ed_user1.timestamp



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, DateTime, ForeignKey
from sqlalchemy.orm import relationship


Base = declarative_base()


class User(Base):
    __tablename__ = "user"
    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    fullname = Column(String(255))
    nickname = Column(String(255))
    timestamp = Column(DateTime)
    addresses = relationship("Adress")


class Adress(Base):
    __tablename__ = "adress"
    id = Column(Integer, primary_key=True)
    title = Column(String(255))
    description = Column(String(255))
    user_id = Column(Integer, ForeignKey("user.id"))
    foos = relationship("Foo")


class Foo(Base):
    __tablename__ = "foo"
    id = Column(Integer, primary_key=True)
    foo = Column(String(255))
    adress_id = Column(Integer, ForeignKey("adress.id"))
    bars = relationship("Bar")


class Bar(Base):
    __tablename__ = "bar"
    id = Column(Integer, primary_key=True)
    bar = Column(String(255))
    foo_id = Column(Integer, ForeignKey("foo.id"))



import datetime
from sqlalchemy import inspect, or_, and_
from sqlalchemy.orm import contains_eager


op_mapping = {
    "eq": lambda attr, value: attr == value,
    "neq": lambda attr, value: attr != value,
    "gt": lambda attr, value: attr > value,
    "gte": lambda attr, value: attr >= value,
    "lt": lambda attr, value: attr < value,
    "lte": lambda attr, value: attr <= value,
}


def generate_query(session, model, raw_query):
    query, _filters = _generate_query(model, session.query(model), raw_query)
    for f in _filters:
        query = query.filter(f)
    return query


def _generate_query(model, query, raw_query, parent=None):
    inst = inspect(model)
    filters = []
    ors = raw_query.get("or")
    ands = raw_query.get("and")
    if ors is not None or ands is not None:
        clauses = ors if ors is not None else ands
        clause_filters = []
        for clause in clauses:
            query, _filters = _generate_query(model, query, clause, parent=parent)
            clause_filters.extend(_filters)
        if ors is not None:
            filters.append(or_(*clause_filters).self_group())
        else:
            filters.append(and_(*clause_filters).self_group())
    for c_attr in inst.mapper.column_attrs:
        prop = c_attr.key
        if prop in raw_query:
            simple_query = raw_query[prop]
            operation = op_mapping[simple_query["op"]]
            value = simple_query["value"]
            if c_attr.columns[0].type.python_type == datetime.datetime and isinstance(
                simple_query["value"], str
            ):
                value = datetime.datetime.fromisoformat(value)
            filters.append(operation(getattr(model, prop), simple_query["value"]))
    for r_attr in inst.mapper.relationships:
        prop = r_attr.key
        if prop in raw_query:
            query = query.join(getattr(model, prop))
            if parent is None:
                parent = contains_eager(getattr(model, prop))
            else:
                parent = parent.contains_eager(getattr(model, prop))
            query = query.options(parent)
            query, _filters = _generate_query(
                r_attr.entity.class_, query, raw_query[prop], parent=parent
            )
            filters.extend(_filters)

    return query, filters
